@@ -1,3 +1,4 @@
+
 #========================================================================
 #
 # PROGRAM FUNCTIONS
@@ -6,6 +7,7 @@
 # Clean-up tmp and lock files
 #
 function cleanup() {
+	[ "$SPINNING" = "off" ] || tput cnorm
 	if [ -e $TMPDIR/error.log ]; then
 	        echo -e "         
 \n==============================================================================
@@ -26,9 +28,162 @@ One or more errors occurred while slackpkg was running:
 }
 trap 'cleanup' 2 14 15 		# trap CTRL+C and kill
 
+# This create an spinning bar
+spinning() {
+	local WAITFILE
+	local SPININTERVAL
+	local COUNT
+	
+	if [ "$SPIN" = "" ]; then
+		SPIN=( "|" "/" "-" "\\" )
+	fi
+	COUNT=${#SPIN[@]}
+		
+	[ -n "$1" ] && WAITFILE=$1 || WAITFILE=/tmp/waitfile
+	[ -n "$2" ] && SPININTERVAL=$2 || SPININTERVAL=0.1
+
+	count=0
+	tput civis 
+	while [ -e $WAITFILE ] ; do 
+		count=$(( count + 1 ))
+		tput sc
+		echo -n ${SPIN[$(( count % COUNT ))]}
+		tput rc
+		sleep $SPININTERVAL
+	done
+	tput cnorm
+}
+
+# System setup
+#
+function system_setup() {
+
+	# Set LOCAL if mirror isn't through network 
+	# If mirror is through network, select the command to fetch
+	# files and packages from there.
+	#
+	MEDIA=${SOURCE%%:*}
+	if [ "$MEDIA" = "cdrom" ] || [ "$MEDIA" = "file" ] || \
+	   [ "$MEDIA" = "local" ]; then
+		SOURCE=/${SOURCE#${MEDIA}://}
+		LOCAL=1
+	else
+		LOCAL=0
+		if [ "$DOWNLOADER" = "curl" ]; then
+			DOWNLOADER="curl ${CURLFLAGS} -o"
+		else
+                	DOWNLOADER="wget ${WGETFLAGS} -O"
+		fi
+	fi
+
+	# Set MORECMD, EDITCMD and check BATCH mode 
+	#
+	if [ "$BATCH" = "on" ] || [ "$BATCH" = "ON" ]; then
+		DIALOG=off
+		SPINNING=off
+		MORECMD=cat
+		EDITCMD=vi
+		if [ "$DEFAULT_ANSWER" = "" ]; then
+			DEFAULT_ANSWER=n
+		fi
+	else
+		if [ "${PAGER}" ]; then	
+			MORECMD="${PAGER}"
+		else
+			MORECMD=more
+		fi
+		if [ "${EDITOR}" ]; then	
+			EDITCMD="${EDITOR}"
+		else
+			EDITCMD=vi
+		fi
+	fi
+
+	# Set ARCH, SLACKKEY and others by slackware port
+	#
+	if [ "$ARCH" = "" ]; then
+		ARCH=$(uname -m)
+	fi
+	case $ARCH in
+		i386|i486|i586|i686)
+			ARCH=[i]*[3456x]86[^_]*
+			SLACKKEY=${SLACKKEY:-"Slackware Linux Project <security@slackware.com>"}
+			PKGMAIN=${PKGMAIN:-slackware}
+		;;
+		x86-64|x86_64|X86-64|X86_64)
+			ARCH=x86[_64]*
+			SLACKKEY=${SLACKKEY:-"Slackware Linux Project <security@slackware.com>"}
+			PKGMAIN=${PKGMAIN:-slackware64}
+		;;
+		s390)
+			ARCH=s390
+			# Slack390 didn't have signed packages
+			CHECKGPG=off
+			PKGMAIN=${PKGMAIN:-slackware}
+		;;
+		arm*)
+			ARCH=arm[el]*
+			SLACKKEY=${SLACKKEY:-"ARMedslack Security (ARMedslack Linux Project Security) <security@armedslack.org>"}
+			PKGMAIN=${PKGMAIN:-slackware}
+		;;
+		powerpc|ppc)
+			ARCH=powerpc
+			SLACKKEY=${SLACKKEY:-"Slackintosh-Project Sign <slackdev@workaround.ch>"}
+			PKGMAIN=${PKGMAIN:-slackintosh}
+		;;
+		*)
+			ARCH=none
+		;;
+	esac
+
+	# Sub %PKGMAIN with the correct $PKGMAIN value
+	#
+	MAIN=$PKGMAIN
+	for i in 0 1 2 3 4 ; do
+		if [ "${PRIORITY[$i]}" = "%PKGMAIN" ]; then
+			PRIORITY[$i]=$PKGMAIN
+		fi
+	done
+
+	TEMPLATEDIR=$CONF/templates
+	if [ ! -d $TEMPLATEDIR ]; then
+	  mkdir $TEMPLATEDIR
+	fi
+
+	SLACKCFVERSION=$(grep "# v[0-9.]\+" $CONF/slackpkg.conf | cut -f2 -dv)
+	CHECKSUMSFILE=$WORKDIR/CHECKSUMS.md5
+	KERNELMD5=$(md5sum /boot/vmlinuz 2>/dev/null)
+	DIALOG_MAXARGS=${DIALOG_MAXARGS:-19500}
+	echo "$0 $VERSION - Slackware Linux $SLACKWARE_VERSION" > $TMPDIR/timestamp
+}
+
 # Syntax Checking
 #
 function system_checkup() {
+
+	# Check slackpkg.conf version
+	#
+	SLCKCFVL=$( expr length $SLACKCFVERSION )
+	if [ "$SLACKCFVERSION" != "$( echo $VERSION |cut -c1-$SLCKCFVL)" ] &&\
+	   [ "$CMD" != "new-config" ]; then
+		echo -e "\
+\nYour slackpkg.conf is outdated. Please, edit it using slackpkg.conf.new\n\
+as example or overwrite it with slackpkg.conf.new.\n\
+\nYou can use 'slackpkg new-config' to do that.\n"
+		cleanup
+	fi
+
+	# Check if ARCH is set
+	#
+	if [ "$ARCH" = "none" ] && [ "$CMD" != "new-config" ]; then
+		echo -e "\
+\nThe ARCH values in slackpkg.conf are now different. You can remove\n\
+ARCH from there, and slackpkg you use your current ARCH or you can look\n\
+at slackpkg.conf.new or slackpkg.conf manual page to see the new valid\n\
+ARCH values\n"
+		cleanup
+	fi
+
 	# Check if the config files are updated to the new slackpkg version
 	#
 	if [ "$WORKDIR" = "" ]; then
@@ -43,7 +198,9 @@ In ${CONF}/slackpkg.conf.new, there is a sample of the new configuration.\n\
 	# Checking if another instance of slackpkg is running
 	#
 	if [ "$(ls /var/lock/slackpkg.* 2>/dev/null)" ] && \
-		[ "$CMD" != "search" ]; then
+		[ "$CMD" != "search" ] && \
+		[ "$CMD" != "help" ] && \
+		[ "$CMD" != "file-search" ]; then
 		echo -e "\
 \nAnother instance of slackpkg is running. If this is not correct, you can\n\
 remove /var/lock/slackpkg.* files and run slackpkg again.\n"
@@ -67,19 +224,27 @@ the problem.\n"
 	# Checking if is the first time running slackpkg
 	#                                               
 	if ! [ -f ${WORKDIR}/pkglist ] && [ "$CMD" != "update" ]; then
-		echo -e "\
+		if [ "$SOURCE" = "" ]; then
+                	echo -e "\
 \nThis appears to be the first time you have run slackpkg.\n\
 Before you install|upgrade|reinstall anything, you need to uncomment\n\
 ONE mirror in ${CONF}/mirrors and run:\n\n\
 \t# slackpkg update\n\n\
 You can see more information about slackpkg functions in slackpkg manpage."
-		cleanup
+			cleanup
+		elif [ "$CMD" != "new-config" ]; then
+			echo -e "\
+\nThe package list is missing.\n\
+Before you install|upgrade|reinstall anything you need to run:\n\n\
+\t# slackpkg update\n"
+			cleanup
+		fi
 	fi                                                      
 
 
 	# Checking if /etc/slackpkg/mirrors are in correct syntax.
 	#                                                         
-	if [ "$SOURCE" = "" ]; then
+	if [ "$SOURCE" = "" ] ; then
 		echo -e "\
 \nYou do not have any mirror selected in ${CONF}/mirrors\n\
 Please edit that file and uncomment ONE mirror.  Slackpkg\n\
@@ -111,6 +276,7 @@ official mirrors can be kept up-to-date.\n"
 	#                                                                    
 	if [ "$(id -u)" != "0" ] && \
 	   [ "$CMD" != "search" ] && \
+	   [ "$CMD" != "file-search" ] && \
 	   [ "$CMD" != "check-updates" ] && \
 	   [ "$CMD" != "info" ]; then
 		echo -e "\n\
@@ -119,15 +285,6 @@ Please log in as root or contact your system administrator.\n"
 		cleanup
 	fi          
 
-	# Check if the mirror are local (cdrom or file)
-	#
-	MEDIA=$(echo ${SOURCE} | cut -f1 -d:)
-	if [ "$MEDIA" = "cdrom" ] || [ "$MEDIA" = "file" ] || \
-	   [ "$MEDIA" = "local" ]; then
-		SOURCE=/$(echo ${SOURCE} | cut -f3- -d/)
-		LOCAL=1
-	fi
-
 	# Check if the "which" command is there
 	if ! which which 1>/dev/null 2>/dev/null ; then
 		echo -e "\n\
@@ -140,10 +297,10 @@ use slackpkg.\n"
 	# md5sum checks
 	#
 	if ! [ $(which md5sum 2>/dev/null) ]; then
-		CHECKPKG=off
+		CHECKMD5=off
 	elif ! [ -f ${WORKDIR}/CHECKSUMS.md5 ] && \
 		[ "$CMD" != "update" ] && \
-		[ "$CHECKPKG" = "on" ]; then
+		[ "$CHECKMD5" = "on" ]; then
 		echo -e "\n\
 No CHECKSUMS.md5 found!  Please disable md5sums checking\n\
 on your ${CONF}/slackpkg.conf or run slackpkg update\n\
@@ -160,6 +317,12 @@ as slackpkg cannot function without awk.\n"
 		cleanup
 	fi
 
+	# Check if tput is there
+	#
+	if ! which tput 1>/dev/null 2>/dev/null ; then
+		SPINNING=off
+	fi
+
 	# Check if gpg is enabled but no GPG command are found.
 	#
 	if ! [ "$(which gpg 2>/dev/null)" ] && [ "${CHECKGPG}" = "on" ]; then
@@ -167,9 +330,9 @@ as slackpkg cannot function without awk.\n"
 		echo -e "\n\
 gpg package not found!  Please disable GPG in ${CONF}/slackpkg.conf or install\n\
 the gnupg package.\n\n\
-To disable GPG, edit slackpkg.conf and change the value of the CHECKGPG variable\n\
-to "off" - you can see an example in the original slackpkg.conf.new file distributed\n\
-with slackpkg.\n"
+To disable GPG, edit slackpkg.conf and change the value of the CHECKGPG \n\
+variable to "off" - you can see an example in the original slackpkg.conf.new\n\
+file distributed with slackpkg.\n"
 		sleep 5
 	fi 
 
@@ -179,7 +342,9 @@ with slackpkg.\n"
 			| grep -c "$SLACKKEY")"
 	if [ "$GPGFIRSTTIME" = "0" ] && \
 		[ "$CMD" != "search" ] && \
+		[ "$CMD" != "file-search" ] && \
 		[ "$CMD" != "info" ] && \
+		[ "$CMD" != "new-config" ] && \
 		[ "$CMD" != "update" ] && \
 		[ "$CMD" != "check-updates" ] && \
 		[ "$CHECKGPG" = "on" ]; then
@@ -188,35 +353,24 @@ You need the GPG key of $SLACKKEY.\n\
 To download and install that key, run:\n\n\
 \t# slackpkg update gpg\n\n\
 You can disable GPG checking too, but it is not a good idea.\n\
-To disable GPG, edit slackpkg.conf and change the value of the CHECKGPG variable\n\
-to "off" - you can see an example in the original slackpkg.conf.new file distributed\n\
-with slackpkg.\n"
+To disable GPG, edit slackpkg.conf and change the value of the CHECKGPG\n\
+variable to "off" - you can see an example in the original slackpkg.conf.new\n\
+file distributed with slackpkg.\n"
 		cleanup
 	fi
-
-	if [ "$BATCH" = "on" ] || [ "$BATCH" = "ON" ]; then
-		DIALOG=off
-		MORECMD=cat
-		if [ "$DEFAULT_ANSWER" = "" ]; then
-			DEFAULT_ANSWER=n
-		fi
-	else
-		MORECMD=more
-	fi
 	echo 
 }
 
 # Got the name of a package, without version-arch-release data
 #
 function cutpkg() {
-	echo ${1/%.t[blxg]z/} | awk -F- 'OFS="-" { 
-				if ( NF > 3 ) { 
-					NF=NF-3
-					print $0 
-				} else {
-					print $0
-				}
-			}'
+	echo ${1/%.t[blxg]z/} | awk -F- -f /usr/libexec/slackpkg/cutpkg.awk
+}
+
+# The same, but reading packages from stdin
+#
+function batchcutpkg() {
+	awk -F- -f /usr/libexec/slackpkg/cutpkg.awk
 }
 
 # Show the slackpkg usage
@@ -224,6 +378,24 @@ function cutpkg() {
 function usage() {
 	echo -e "\
 slackpkg - version $VERSION\n\
+\nUsage:\n\
+\tslackpkg [OPTIONS] {install|remove|search|file-search|
+\t\t\t    upgrade|reinstall|blacklist} {PATTERN|FILE}
+\tslackpkg [OPTIONS] {generate-template|install-template|remove-template}
+\t\t\t   TEMPLATENAME
+\tslackpkg [OPTIONS] info PACKAGE
+\tslackpkg [OPTIONS] update [gpg]
+\tslackpkg [OPTIONS] {clean-system|upgrade-all|install-new}
+\tslackpkg [OPTIONS] {new-config|check-updates|help}
+\nIf you need more information try to use 'slackpkg help' or look the\n\
+slackpkg's manpage.
+"
+	cleanup
+}
+
+function full_usage() {
+	echo -e "\
+slackpkg - version $VERSION\n\
 \nUsage: \tslackpkg update [gpg]\t\tdownload and update files and 
 \t\t\t\t\tpackage indexes
 \tslackpkg check-updates\t\tcheck if there is any news on
@@ -251,18 +423,61 @@ slackpkg - version $VERSION\n\
 \tslackpkg download\t\tOnly download (do not install) a package
 \tslackpkg info package\t\tShow package information 
 \t\t\t\t\t(works with only ONE package)
-\tslackpkg search file\t\tSearch for a specific file in the
+\tslackpkg search package\t\tSearch packages that have a
+\t\t\t\t\tselected name
+\tslackpkg file-search file\tSearch for a specific file in the
 \t\t\t\t\tentire package collection
 \tslackpkg new-config\t\tSearch for new configuration files and
 \t\t\t\t\task to user what to do with them.
+\tslackpkg generate-template\tCreate a template with all 
+\t\t\t\t\tofficial Slackware packages installed 
+\t\t\t\t\tin your machine.
+\tslackpkg install-template\tInstall selected template.
+\tslackpkg remove-template\tRemove selected template. Be careful.
+\tslackpkg help\t\t\tShow this screen. 
 \nYou can see more information about slackpkg usage and some examples
-in slackpkg's manpage. You can use partial package names (such as x11
-instead x11-devel, x11-docs, etc), or even Slackware series
+in slackpkg's manpage. You can use partial package names (such as xorg 
+instead of xorg-server, xorg-docs, etc), or even Slackware series
 (such as "n","ap","xap",etc) when searching for packages.
 "
 	cleanup
 }
 
+# Verify if we have enough disk space to install selected package
+#
+function havespace() {
+	local DSIZE
+	local ASIZE
+	DSIZE=$(grep "^${1}" ${TMPDIR}/tempsize | \
+		awk 'BEGIN { tot=0 } { tot+=$2 } END { print int(tot/1024)+1}')
+	ASIZE=$(df ${1} | awk '/% \// { print 0+$(NF-2) }')
+	if [ ${DSIZE} -gt ${ASIZE} ] ; then
+		ISOK=0
+	fi
+}
+
+function checksize() {
+	local i
+	local ISOK=1
+	tar -tvf ${1} | tr -s ' ' | grep -v '^[dl]' | cut -f6,3 -d\ | \
+	sed 's,[^/]*$,,' | awk '
+	{ size[$2]+=$1 }
+	END { 
+	    	for (i in size) { 
+			print "/"i,size[i]
+		}
+	}' > ${TMPDIR}/tempsize
+
+	for i in $(tac /proc/mounts | grep "^/dev" |cut -f2 -d\ ); do
+		if grep -q "^${i}" ${TMPDIR}/tempsize ; then
+			havespace ${i}
+			grep -v "^${i}/" ${TMPDIR}/tempsize > ${TMPDIR}/tempsize.tmp
+			mv ${TMPDIR}/tempsize.tmp ${TMPDIR}/tempsize
+		fi	
+	done
+	echo ${ISOK}
+}
+
 # Verify if the package was corrupted by checking md5sum
 #
 function checkmd5() {
@@ -279,11 +494,14 @@ function checkmd5() {
 	fi
 }
 
+# Verify the GPG signature of files/packages
+#
 function checkgpg() {
 	gpg --verify ${1}.asc ${1} 2>/dev/null && echo "1" || echo "0"
 }
 
-
+# Check if package blacklisted 
+#
 function checkblacklist {
         local i
         local BLKNAME="${PKGDATA[1]}"
@@ -301,6 +519,10 @@ function checkblacklist {
 	fi
 }
 
+# Found packages in repository. 
+# This function selects the package from the higher priority
+# reposistory directories.
+#
 function givepriority {
         local DIR
         local ARGUMENT=$1
@@ -310,9 +532,9 @@ function givepriority {
         unset FULLNAME
 	unset PKGDATA
 	
-        for DIR in $FIRST $SECOND $THIRD $FOURTH $FIFTH ; do
+        for DIR in ${PRIORITY[@]} ; do
 		[ "$PKGDATA" ] && break
-                PKGDATA=( $(grep "^${DIR} ${ARGUMENT} " ${WORKDIR}/pkglist) )
+                PKGDATA=( $(grep "^${DIR} ${ARGUMENT} " ${TMPDIR}/pkglist) )
                 if [ "$PKGDATA" ]; then
                         checkblacklist
                         if [ "$?" = "1" ]; then
@@ -327,6 +549,16 @@ function givepriority {
         done
 }
 
+# Creates files with mirror package names (spkg), local package
+# names (lpkg) and packages unique to one or other file (dpkg)
+#
+function listpkgname() {
+	cut -f2 -d\  ${TMPDIR}/pkglist | sort > ${TMPDIR}/spkg	
+	cut -f2 -d\  ${TMPDIR}/tmplist | sort > ${TMPDIR}/lpkg
+	cat ${TMPDIR}/pkglist ${TMPDIR}/tmplist | \
+		cut -f2-6 -d\ |sort | uniq -u | \
+		cut -f1 -d\  | uniq > ${TMPDIR}/dpkg
+}
 
 # Function to make install/reinstall/upgrade lists
 #
@@ -338,6 +570,9 @@ function makelist() {
 	INPUTLIST=$@
 
 	ls -1 /var/log/packages/* | awk -f /usr/libexec/slackpkg/pkglist.awk > ${TMPDIR}/tmplist
+	cp ${WORKDIR}/pkglist ${TMPDIR}/pkglist
+
+	touch ${TMPDIR}/waiting
 
 	case "$CMD" in
 		clean-system)
@@ -349,23 +584,28 @@ function makelist() {
 		install-new)
 			echo -n "Looking for NEW packages to install. Please wait... "
 		;;
+		*-template)
+			echo -n "Looking for packages in \"$ARG\" template to ${CMD/%-template/}. Please wait..."
+		;;
 		*)
 			echo -n "Looking for $(echo $INPUTLIST | tr -d '\\') in package list. Please wait... "
 		;;
 	esac
 
+	[ "$SPINNING" = "off" ] || spinning ${TMPDIR}/waiting &
+
 	case "$CMD" in
 		download)
 			for ARGUMENT in $(echo $INPUTLIST); do
-				for i in $(grep -w -- "${ARGUMENT}" ${WORKDIR}/pkglist | cut -f2 -d\  | sort -u); do
-					LIST="$LIST $(grep " ${i} " ${WORKDIR}/pkglist | cut -f6,8 -d\  --output-delimiter=.)"
+				for i in $(grep -w -- "${ARGUMENT}" ${TMPDIR}/pkglist | cut -f2 -d\  | sort -u); do
+					LIST="$LIST $(grep " ${i} " ${TMPDIR}/pkglist | cut -f6,8 -d\  --output-delimiter=.)"
 				done
 				LIST="$(echo -e $LIST | sort -u)"
 			done
 		;;
 		blacklist)
 			for ARGUMENT in $(echo $INPUTLIST); do
-				for i in $(cat ${WORKDIR}/pkglist ${TMPDIR}/tmplist | \
+				for i in $(cat ${TMPDIR}/pkglist ${TMPDIR}/tmplist | \
 						grep -w -- "${ARGUMENT}" | cut -f2 -d\  | sort -u); do
 					grep -qx "${i}" ${CONF}/blacklist || LIST="$LIST $i"
 				done
@@ -373,7 +613,7 @@ function makelist() {
 		;;
 		install|upgrade|reinstall)
 			for ARGUMENT in $(echo $INPUTLIST); do
-				for i in $(grep -w -- "${ARGUMENT}" ${WORKDIR}/pkglist | cut -f2 -d\  | sort -u); do
+				for i in $(grep -w -- "${ARGUMENT}" ${TMPDIR}/pkglist | cut -f2 -d\  | sort -u); do
 					givepriority $i
 					[ ! "$FULLNAME" ] && continue
 
@@ -399,7 +639,7 @@ function makelist() {
 		;;
 		remove)
 			for ARGUMENT in $(echo $INPUTLIST); do
-				for i in $(cat ${WORKDIR}/pkglist ${TMPDIR}/tmplist | \
+				for i in $(cat ${TMPDIR}/pkglist ${TMPDIR}/tmplist | \
 					  	grep -w -- "${ARGUMENT}" | cut -f6 -d\  | sort -u); do
 					PKGDATA=( $(grep -w -- "$i" ${TMPDIR}/tmplist) )
 					[ ! "$PKGDATA" ] && continue
@@ -410,17 +650,20 @@ function makelist() {
 			done
 		;;
 		clean-system)
-	                for i in $(cut -f6 -d\  ${TMPDIR}/tmplist); do
-				NAME=$(cutpkg $i)
-				if [ $(cut -f2 -d\  ${WORKDIR}/pkglist |grep -cx "${NAME}") = "0" ] &&
-					[ $(grep -cx "${NAME}" $CONF/blacklist) = 0 ]; then
-					LIST="$LIST $i"
-                        	fi
-                	done
+			listpkgname
+			for i in $(comm -2 -3 ${TMPDIR}/lpkg ${TMPDIR}/spkg) ; do
+				PKGDATA=( $(grep -w -- "$i" ${TMPDIR}/tmplist) )
+				[ ! "$PKGDATA" ] && continue
+				checkblacklist && continue
+				LIST="$LIST ${PKGDATA[5]}" 
+				unset PKGDATA
+			done				
 		;;
 		upgrade-all)
-			for i in $(cut -f2 -d\  ${TMPDIR}/tmplist); do
-
+			listpkgname
+			for i in $(comm -1 -2 ${TMPDIR}/lpkg ${TMPDIR}/dpkg | \
+				   comm -1 -2 - ${TMPDIR}/spkg) ; do
+	
 				givepriority ${i}
 				[ ! "$FULLNAME" ] && continue
 
@@ -443,8 +686,57 @@ function makelist() {
 					LIST="$LIST ${FULLNAME}"
 			done
 		;;
+		install-template)
+			for i in $INPUTLIST ; do
+				givepriority $i
+				[ ! "$FULLNAME" ] && continue
+				grep -q " ${NAME} " ${TMPDIR}/tmplist || \
+					LIST="$LIST ${FULLNAME}"
+			done
+		;;	
+		remove-template)
+			for i in $INPUTLIST ; do
+				givepriority $i
+				[ ! "$FULLNAME" ] && continue
+				grep -q " ${NAME} " ${TMPDIR}/tmplist && \
+					LIST="$LIST ${FULLNAME}"
+			done
+		;;	
+		search|file-search)
+			if [ "$CMD" = "file-search" ]; then
+				# Search filelist.gz for possible matches
+				for i in ${PRIORITY[@]}; do
+					if [ -e ${WORKDIR}/${i}-filelist.gz ]; then
+						PKGS="$(zegrep -w "${INPUTLIST}" ${WORKDIR}/${i}-filelist.gz | \
+							cut -d\  -f 1 | awk -F'/' '{print $NF}')"
+						for FULLNAME in $PKGS ; do
+							NAME=$(cutpkg ${FULLNAME})
+							echo $LIST | \
+								grep -qe "${NAME}-[^-]\+-\(${ARCH}\|fw\|noarch\)-[^-]\+" && \
+								continue
+							LIST="$LIST ${FULLNAME}"
+						done
+					fi
+				done
+			else
+				for i in ${PRIORITY[@]}; do
+					PKGS=$(grep "^${i}.*${PATTERN}" \
+						${TMPDIR}/pkglist | cut -f6 -d\ )
+					for FULLNAME in $PKGS ; do
+						NAME=$(cutpkg ${FULLNAME})
+						echo $LIST | \
+							grep -qe "${NAME}-[^-]\+-\(${ARCH}\|fw\|noarch\)-[^-]\+" && \
+							continue
+						LIST="$LIST ${FULLNAME}"
+					done
+				done
+			fi
+		;;	
 	esac
 	LIST=$(echo -e $LIST | tr \  "\n" | uniq )
+
+	rm ${TMPDIR}/waiting
+
 	echo -e "DONE\n"
 }
 
@@ -467,6 +759,49 @@ function answer() {
 	fi
 }
 
+function searchlist() {
+	local i
+	local BASENAME
+	local RAWNAME
+	local STATUS
+	local INSTPKG
+
+	for i in $1; do
+	    if [ "$BASENAME" = "$(cutpkg ${i})" ]; then
+		continue
+	    fi
+	    # BASENAME is base package name 
+	    BASENAME="$(cutpkg ${i})"
+
+	    # RAWNAME is Latest available version  
+	    RAWNAME="${i/%.t[blxg]z/}"
+
+	    # Default is uninstalled
+	    STATUS="uninstalled"
+
+	    # First is the package already installed?
+	    # Amazing what a little sleep will do
+	    # exclusion is so much nicer :)
+	    INSTPKG=$(ls -1 /var/log/packages | \
+		grep -e "^${BASENAME}-[^-]\+-\(${ARCH}\|fw\|noarch\)-[^-]\+")
+
+		# INSTPKG is local version
+		if [ ! "${INSTPKG}" = "" ]; then
+
+		# If installed is it uptodate?
+		if [ "${INSTPKG}" = "${RAWNAME}" ]; then
+		    STATUS=" installed "
+		    echo "[${STATUS}] - ${INSTPKG}"
+		else
+		    STATUS="  upgrade  "
+		echo "[${STATUS}] - ${INSTPKG} --> ${RAWNAME}"
+		fi
+		else
+		    echo "[${STATUS}] - ${RAWNAME}"
+		fi
+	done
+}
+
 # Show the lists and asks if the user want to proceed with that action
 # Return accepted list in $SHOWLIST
 #
@@ -489,11 +824,15 @@ function showlist() {
 
 function getfile() {
         if [ "$LOCAL" = "1" ]; then
-                echo -e "\t\t\tCopying $1..."
-                cp ${SOURCE}$1 $2 2>/dev/null
+                echo -e "\t\t\tLinking $1..."
+                if [ -e $1 ]; then
+			ln -s $1 $2 2>/dev/null
+		else
+			return 1
+		fi
         else
                 echo -e "\t\t\tDownloading $1..."
-                wget ${WGETFLAGS} ${SOURCE}$1 -O $2
+		$DOWNLOADER $2 $1
         fi
 }                                                       
 
@@ -507,7 +846,7 @@ function getpkg() {
 	local NAMEPKG
 	local CACHEPATH
 
-	PKGNAME=( $(grep -w -m 1 -- "${1/%.t[blxg]z/}" ${WORKDIR}/pkglist) )
+	PKGNAME=( $(grep -w -m 1 -- "${1/%.t[blxg]z/}" ${TMPDIR}/pkglist) )
 	NAMEPKG=${PKGNAME[5]}.${PKGNAME[7]}
 	FULLPATH=${PKGNAME[6]}
 	CACHEPATH=${TEMP}/${FULLPATH}
@@ -523,21 +862,14 @@ function getpkg() {
 		# to CACHEPATH else, download packages from remote host and 
 		# put then in CACHEPATH
 		#
-		if [ "${LOCAL}" = "1" ]; then 
-                	echo -e "\tCopying $NAMEPKG..."
-			cp ${SOURCE}${FULLPATH}/${NAMEPKG} ${CACHEPATH}
-			if [ "$CHECKGPG" = "on" ]; then
-				cp ${SOURCE}${FULLPATH}/${NAMEPKG}.asc ${CACHEPATH}
-			fi
-		else
-                	echo -e "\tDownloading $NAMEPKG..."
-			wget ${WGETFLAGS} -P ${CACHEPATH} -nd ${SOURCE}${FULLPATH}/${NAMEPKG}
-			if [ "$CHECKGPG" = "on" ]; then
-				wget ${WGETFLAGS} -P ${CACHEPATH} -nd ${SOURCE}${FULLPATH}/${NAMEPKG}.asc
-			fi
+		getfile ${SOURCE}${FULLPATH}/${NAMEPKG} \
+			${CACHEPATH}/${NAMEPKG} 
+		if [ "$CHECKGPG" = "on" ]; then
+			getfile ${SOURCE}${FULLPATH}/${NAMEPKG}.asc \
+				${CACHEPATH}/${NAMEPKG}.asc
 		fi
 
-		if ! [ -e $CACHEPATH/$1 ]; then
+		if ! [ -e ${CACHEPATH}/$1 ]; then
 			ERROR="Not found"
 			ISOK="0"
 			echo -e "${NAMEPKG}:\t$ERROR" >> $TMPDIR/error.log
@@ -546,10 +878,19 @@ function getpkg() {
 		echo -e "\tPackage $1 is already in cache - not downloading" 
 	fi
 
+	# Check if we have sufficient disk space to install selected package
+        if [ "$CHECKSIZE" = "on" ] && [ "$ISOK" = "1" ]; then
+		ISOK=$(checksize ${CACHEPATH}/$1)
+		if [ "$ISOK" = "0" ]; then
+			ERROR="Insufficient disk space"
+			echo -e "${NAMEPKG}:\t$ERROR" >> $TMPDIR/error.log
+		fi
+	fi
+
 	# If MD5SUM checks are enabled in slackpkg.conf, check the
 	# packages md5sum to detect if they are corrupt or not
 	#
-	if [ "$CHECKPKG" = "on" ] && [ "$ISOK" = "1" ]; then
+	if [ "$CHECKMD5" = "on" ] && [ "$ISOK" = "1" ]; then
 		ISOK=$(checkmd5 ${CACHEPATH}/$1)
 		if [ "$ISOK" = "0" ]; then 
 			ERROR="md5sum"
@@ -615,7 +956,7 @@ function checkchangelog()
 	# Download ChangeLog.txt first of all and test if it's equal
 	# or different from our already existent ChangeLog.txt 
 	#
-	getfile ChangeLog.txt $TMPDIR/ChangeLog.txt
+	getfile ${SOURCE}ChangeLog.txt $TMPDIR/ChangeLog.txt
 	if ! grep -q "[a-z]" $TMPDIR/ChangeLog.txt ; then
 		echo -e "\
 \nError downloading from $SOURCE.\n\
@@ -642,7 +983,7 @@ function updatefilelists()
 		fi
 	fi
 	echo
-	cp $TMPDIR/ChangeLog.txt ${WORKDIR}/ChangeLog.txt
+	cp ${TMPDIR}/ChangeLog.txt ${WORKDIR}/ChangeLog.txt
 
 	#
 	# Download MANIFEST, FILELIST.TXT and CHECKSUMS.md5
@@ -651,15 +992,15 @@ function updatefilelists()
 	# That will be download MANIFEST.bz2 files
 	#
 	echo -e "\t\tList of all files"
-	for i in $FIRST $SECOND $THIRD $FOURTH $FIFTH ; do 
-		getfile ${i}/MANIFEST.bz2 $TMPDIR/${i}-MANIFEST.bz2 && \
+	for i in ${PRIORITY[@]} ; do 
+		getfile ${SOURCE}${i}/MANIFEST.bz2 $TMPDIR/${i}-MANIFEST.bz2 && \
 			DIRS="$DIRS $i"
 	done
 
 	ISOK="1"
 	echo -e "\t\tChecksums"
-	getfile CHECKSUMS.md5 ${TMPDIR}/CHECKSUMS.md5
-	getfile CHECKSUMS.md5.asc ${TMPDIR}/CHECKSUMS.md5.asc
+	getfile ${SOURCE}CHECKSUMS.md5 ${TMPDIR}/CHECKSUMS.md5
+	getfile ${SOURCE}CHECKSUMS.md5.asc ${TMPDIR}/CHECKSUMS.md5.asc
 	if ! [ -e "${TMPDIR}/CHECKSUMS.md5" ]; then
 		echo -e "\
 \n\t\tWARNING: Your mirror appears incomplete and is missing the\n\
@@ -679,7 +1020,7 @@ function updatefilelists()
 \t\t       or has been tampered with.\n"
 				cleanup
 			fi
-		else
+		elif [ "$SLACKKEY" != "" ]; then
 			echo -e "\
 \n\t\tWARNING: Without CHECKGPG, we can't check if this file is\n\
 \t\t         signed by:\n\
@@ -692,10 +1033,10 @@ function updatefilelists()
 
 	ISOK="1"
 	echo -e "\t\tPackage List"
-	getfile FILELIST.TXT $TMPDIR/FILELIST.TXT
-	if [ "$CHECKPKG" = "on" ]; then
-		CHECKSUMSFILE=$TMPDIR/CHECKSUMS.md5
-		ISOK=$(checkmd5 $TMPDIR/FILELIST.TXT)
+	getfile ${SOURCE}FILELIST.TXT ${TMPDIR}/FILELIST.TXT
+	if [ "$CHECKMD5" = "on" ]; then
+		CHECKSUMSFILE=${TMPDIR}/CHECKSUMS.md5
+		ISOK=$(checkmd5 ${TMPDIR}/FILELIST.TXT)
 	fi
 	if [ "$ISOK" = "1" ]; then 
 		if ! [ -e $WORKDIR/LASTUPDATE ]; then
@@ -742,7 +1083,7 @@ function updatefilelists()
 	# 
 	echo -e "\t\tPackage descriptions"
 	for i in $DIRS; do
-		getfile ${i}/PACKAGES.TXT $TMPDIR/${i}-PACKAGES.TXT
+		getfile ${SOURCE}${i}/PACKAGES.TXT $TMPDIR/${i}-PACKAGES.TXT
 	done
 
 	# Format FILELIST.TXT
@@ -803,15 +1144,20 @@ function sanity_check() {
 	local DOUBLEFILES
 	local ANSWER
 
+	touch ${TMPDIR}/waiting
+	echo -e "Checking local integrity... \c"
+
+	[ "$SPINNING" = "off" ] || spinning ${TMPDIR}/waiting &
+
 	for i in $(ls -1 /var/log/packages | \
 		egrep -- "^.*-(${ARCH}|fw|noarch)-[^-]+-upgraded"); do
 		REVNAME=$(echo ${i} | awk -F'-upgraded' '{ print $1 }')
 		mv /var/log/packages/${i} /var/log/packages/${REVNAME}
 		mv /var/log/scripts/${i} /var/log/scripts/${REVNAME}
-	done 
-	for i in $(ls -1 /var/log/packages | egrep "^.*-(${ARCH}|fw|noarch)-[^-]+$"); do
-		cutpkg $i
-	done | sort > $TMPDIR/list1
+	done
+	
+	ls -1 /var/log/packages | egrep "^.*-(${ARCH}|fw|noarch)-[^-]+$" | \
+				  batchcutpkg | sort > $TMPDIR/list1 
 	cat $TMPDIR/list1 | uniq > $TMPDIR/list2
 	FILES="$(diff $TMPDIR/list1 $TMPDIR/list2 | grep '<' | cut -f2 -d\ )"
 	if [ "$FILES" != "" ]; then
@@ -821,6 +1167,10 @@ function sanity_check() {
 		done
 		unset FILES
 	fi
+
+	rm ${TMPDIR}/waiting
+	echo -e "DONE"
+
 	if [ "$DOUBLEFILES" != "" ]; then
 		echo -e "\
 You have a broken /var/log/packages - with two versions of the same package.\n\
@@ -907,3 +1257,98 @@ function install_pkg() {
 		getpkg $i installpkg Installing
 	done
 }
+
+#
+# Template related functions
+#
+include_includes() {
+	TEMPLATEFILE=$1
+	COUNT=$((COUNT + 1))
+	TMPFILE=$TMPDIR/$(basename $( echo $TEMPLATEFILE | cut -f1,2 -d. )).$COUNT.tmp
+	INCLUDELIST="$( grep "^#include" $TEMPLATEFILE | cut -d\  -f2 )"
+	if [ "$INCLUDELIST" != "" ]; then
+		for INCLUDE in $INCLUDELIST ; do
+			echo "#include $INCLUDE" \
+				>> ${TMPFILE/.$COUNT/}.header
+			cat $INCLUDE > $TMPFILE
+			grep -v "^$" $TEMPLATEFILE | grep -v "^#" >> $TMPFILE
+			include_includes $TMPFILE
+		done
+	else
+		echo $TEMPLATEFILE
+		return
+	fi
+}
+
+parse_template() {
+	if [ "$USE_INCLUDES" = "off" ]; then
+		touch $TMPDIR/$1.header
+		grep -v "^$" $1 | grep -v "^#" | sort -u > $TMPDIR/$1.tmp
+	else
+		TMPFILE=$(include_includes $1)
+		sort -u $TMPFILE > $TMPDIR/$1.tmp
+		if [ -e $TMPDIR/$1.tmp.header ]; then
+			sort -u $TMPDIR/$1.tmp.header > $TMPDIR/$1.header
+		fi
+		rm $TMPDIR/$1.[0-9]*.tmp $TMPDIR/$1.tmp.header 2>/dev/null
+	fi
+}
+
+generate_template() {
+	if [ "$USE_INCLUDES" = "on" ]; then
+		(
+			cd $TEMPLATEDIR
+			if [ "$(ls *.template 2>/dev/null)" != "" ]; then
+				echo -e "\tParsing actual template files:"
+				for i in *.template ; do
+					echo -e "\t\t$i"
+					parse_template $i
+				done
+			fi
+		)
+	fi
+
+	touch $TMPDIR/allheaders
+
+	touch $TMPDIR/waiting
+	echo -e "\tGenerating slackware installed package list (this may take a while) \c"
+	[ "$SPINNING" = "off" ] || spinning ${TMPDIR}/waiting &
+	for i in /var/log/packages/* ; do 
+		PKGNAME=$( cutpkg $(basename $i))
+		grep -q " $PKGNAME " ${WORKDIR}/pkglist && \
+			echo $PKGNAME >> $TMPDIR/$TEMPLATE.work
+	done  
+	rm $TMPDIR/waiting
+	echo " "
+
+	echo -e "\tGenerating $TEMPLATE "
+	for TMPLATE in $( wc -l $TMPDIR/* | sort -r | \
+			  awk -F/ '/template.tmp/ { print $NF }'); do
+		if ! $( grep -q "^#include.*${TMPLATE/.tmp/}" \
+			$TMPDIR/allheaders) ; then
+			diff -y $TMPDIR/$TEMPLATE.work $TMPDIR/$TMPLATE | \
+				awk -vTMPDIR=$TMPDIR \
+				 	'!/</ { print $1 > TMPDIR"/same" } 
+					 /</ { print $1 > TMPDIR"/notsame" }'
+			if $( diff -q 	$TMPDIR/$TMPLATE \
+					$TMPDIR/same &>/dev/null ); then
+				echo "#include ${TMPLATE/.tmp/}" \
+					>> $TMPDIR/$TEMPLATE.header
+				cat $TMPDIR/${TMPLATE/.tmp/}.header \
+					>> $TMPDIR/allheaders 2>/dev/null
+				cat $TMPDIR/same >> $TMPDIR/allfiles
+			fi
+		fi
+	done
+
+	if [ -e $TMPDIR/allfiles ]; then
+		sort -u $TMPDIR/allfiles > $TMPDIR/alluniqfiles
+	else
+		touch $TMPDIR/alluniqfiles
+	fi
+	if [ -e $TMPDIR/$TEMPLATE.header ]; then
+		cat $TMPDIR/$TEMPLATE.header > $TEMPLATEDIR/$TEMPLATE
+	fi
+	diff $TMPDIR/alluniqfiles $TMPDIR/$TEMPLATE.work |\
+		 awk '/>/ { print $2 }' >> $TEMPLATEDIR/$TEMPLATE
+}
